"""
setops_play.py -- Play around with set operations

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relation import Relation
from pyral.rtypes import Attribute, SetOp


sdb = "sdb"  # Set operation db example

Element_i = namedtuple('Element_i', 'Name')

class SetPlay:
    """
    SetOp examples
    """

    @classmethod
    def setup(cls):
        """
        Initialize the examples
        """
        Database.open_session(name=sdb)
        Relation.create(db=sdb, attrs=[Attribute(name="Name", type="string")],
                        tuples=[
                            # Element_i(Name="A"),
                            Element_i(Name="B"),
                            Element_i(Name="C"),
                        ], svar_name="set2")
        Relation.create(db=sdb, attrs=[Attribute(name="Name", type="string")],
                        tuples=[
                            # Element_i(Name="A"),
                            Element_i(Name="B"),
                            Element_i(Name="C"),
                        ])

        x = Relation.set_compare(db=sdb, rname2="set2", op=SetOp.psuperset)
        pass


""" test_dogs.py -- Test commands using the TclRAL manpage dog examples """

import pytest
from pyral.relation import Relation
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.rtypes import Attribute, RelationValue

from collections import namedtuple

DOG_i = namedtuple('DOG_i', 'DogName')
OWNER_i = namedtuple('OWNER_i', 'OwnerName')
OWNERSHIP_i = namedtuple('OWNERSHIP_i', 'OwnerName DogName')
db= "dog_db"

@pytest.fixture(scope='module')
def open_db_session():
    Database.open_session(db)
    yield
    # don't close here — it's handled in tear_down

@pytest.fixture(autouse=True)
def tear_down(open_db_session):
    yield
    Database.close_session(db)

@pytest.fixture(scope='module')
def dog_db(open_db_session):
    Relvar.create_relvar(db=db, name="DOG", attrs=[
        Attribute(name="DogName", type="string")
    ], ids={1: ["DogName"]})
    Relvar.create_relvar(db=db, name="OWNER", attrs=[
        Attribute(name="OwnerName", type="string"),
        # Attribute(name="Age", type="int"),
        # Attribute(name="City", type="string"),
    ], ids={1: ["OwnerName"]})
    Relvar.create_relvar(db=db, name="OWNERSHIP", attrs=[
        Attribute(name="DogName", type="string"),
        Attribute(name="OwnerName", type="string")
    ], ids={1: ["OwnerName", "DogName"]})
    Relvar.insert(db=db, relvar='DOG', tuples=[
        DOG_i(DogName="Fido"),
        DOG_i(DogName="Sam"),
        DOG_i(DogName="Spot"),
        DOG_i(DogName="Rover"),
        DOG_i(DogName="Fred"),
        DOG_i(DogName="Jumper"),
    ])
    Relvar.insert(db=db, relvar='OWNER', tuples=[
        OWNER_i(OwnerName="Sue"),
        OWNER_i(OwnerName="George"),
        OWNER_i(OwnerName="Alice"),
        OWNER_i(OwnerName="Mike"),
        OWNER_i(OwnerName="Jim"),
    ])
    Relvar.insert(db=db, relvar='OWNERSHIP', tuples=[
        OWNERSHIP_i(OwnerName="Sue", DogName="Fido"),
        OWNERSHIP_i(OwnerName="Sue", DogName="Sam"),
        OWNERSHIP_i(OwnerName="George", DogName="Fido"),
        # OWNERSHIP_i(OwnerName="George", DogName="Sam"),
        OWNERSHIP_i(OwnerName="Alice", DogName="Spot"),
        OWNERSHIP_i(OwnerName="Mike", DogName="Rover"),
        OWNERSHIP_i(OwnerName="Jim", DogName="Fred"),
    ])


def test_divide(dog_db):
    # Dividend
    Relation.project(db=db, relation="DOG", attributes=("DogName",), svar_name="dividend")

    # Divisor
    R = f"OwnerName:<{'George'}> OR OwnerName:<{'Sue'}>"
    Relation.restrict(db=db, relation='OWNER', restriction=R)
    Relation.project(db=db, attributes=("OwnerName",), svar_name="divisor")

    # Mediator
    Relation.project(db=db, relation="OWNERSHIP", attributes=("OwnerName", "DogName"), svar_name="mediator")
    result = Relation.divide(db=db, dividend="dividend", divisor="divisor", mediator="mediator", svar_name="quotient")

    expected = RelationValue(name='^relation', header={'DogName': 'string'}, body=[{'DogName': 'Fido'}])
    assert result == expected




"""
restrict_play.py -- Test advanced restriction

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute, Order, Card, Extent

Shaft_i = namedtuple('Shaft_i', 'ID Speed Name In_service')

ev = "ev"


def play():
    Database.open_session(ev)
    Relvar.create_relvar(
        db=ev, name="Shaft", attrs=[
            Attribute(name="ID", type="string"),
            Attribute(name="Speed", type="double"),
            Attribute(name="Name", type="string"),
            Attribute(name="In_service", type="boolean")], ids={1: ["ID"]}
    )
    Relvar.insert(db=ev, relvar="Shaft", tuples=[
                        Shaft_i(ID='S1', Speed=31.3, Name="hello", In_service=True),
                        Shaft_i(ID='S2', Speed=14.2, Name="hi there", In_service=False),
                        Shaft_i(ID='S3', Speed=20.16, Name="NOT here", In_service=True),
                        Shaft_i(ID='S4', Speed=31.3, Name="A", In_service=True),
                    ]
    )
    n = "NOT here"
    i = "S3"
    R = f"Name:<{n}>, ID:<{i}>"
    result = Relation.restrict(db=ev, relation="Shaft", restriction=R, svar_name="s_rv")
    Relation.print(db=ev, variable_name="s_rv")
    pass
    # Relvar.printall(db=ev)
    # Relation.create(db=ev, attrs=[
    #     Attribute(name="ID", type="string"),
    #     Attribute(name="Speed", type="double"),
    #     Attribute(name="In_service", type="boolean")],
    #                 tuples=[
    #                     Shaft_i(ID='S1', Speed=31.3, In_service=False),
    #                     Shaft_i(ID='S2', Speed=14.2, In_service=False),
    #                     Shaft_i(ID='S3', Speed=20.16, In_service=False),
    #                 ], svar_name="shafts_rv")
    #
    # Relvar.set(db=ev, relvar="Shaft", relation="shafts_rv")
    # Relation.set()
    # Relation.raw(db=ev, cmd_str="relation restrictwith ${shafts_rv} {[expr {$Speed > 14}]}",
    #              svar_name="raw_rv")
    # Relation.raw(db=ev, cmd_str="relation restrictwith ${shafts_rv} {[string match S* {S2}]}",
    #              svar_name="raw_rv")
    # Relation.raw(db=ev, cmd_str="relation restrict ${shafts_rv} t {[string match {S2} [tuple extract $t ID]]}",
    #              svar_name="raw_rv")
    # Relation.print(db=ev, variable_name="raw_rv")
    pass

    # Relation.raw(db=ev, cmd_str=r"relation restrictwith ${shafts_rv} [ [string match {S1} $ID] && [string match {True} $In_service] ]",
    #              svar_name="raw_rv")
    # R = f"In_service:<{True}>"
    # R = f"ID:<{v}>, In_service:<{True}>"
    # R = f"ID:<{v}> OR In_service:<{True}>"

    v = "S1"
    s = 14.2
    # R = f"Speed > {s}, ID:<S3>"
    # R = f"Class:{v}"
    # R = f"ID==<S1>"
    # R = f"ID:<{v}>, In_service:<True>"
    # R = f"ID:<{v}>, Speed:<{s}>"
    # R = f"In_service:<{True}>"
    # Not working: R = f"ID:<{v}>, Speed:{s}"
    # R = f"Speed:{s}"
    R = f"Speed >= {s}"
    # R = f"NOT In_service:<{True}>"
    # R = f"NOT Speed:<{s}>"
    # s = 31
    # R = f"ID:<{v}> OR (In_service:<{True}> AND Speed > {s})"

    # result = Relation.restrict(db=ev, relation="shafts_rv", svar_name="restriction")
    # result = Relation.restrict(db=ev, relation="shafts_rv", restriction=R, svar_name="restriction")
    # Relation.print(db=ev, variable_name="restriction")

    # result = Relation.rank(db=ev, relation="shafts_rv", sort_attr_name="Speed", order=Order.DESCENDING)
    # result = Relation.rank_restrict(db=ev, relation="shafts_rv", attr_name="Speed", extent=Extent.GREATEST,
    #                                 card=Card.ONE, svar_name="rr_result")
    # Relation.tag(db=ev, relation="Shaft", tag_attr_name="_instance", svar_name="shaft_tagged")
    # Relation.project(db=ev, relation="shaft_tagged", attributes=("ID", "_instance"), svar_name="shaft_ids")
    # Relation.print(db=ev, variable_name="shaft_ids")
    Relvar.printall(db=ev)
    # Relation.print(db=ev, variable_name="Shaft")

    # pass

    # Relvar.create_relvar(db=ev, name='Shaft', attrs=[Attribute('ID', 'string'), Attribute('In_service', 'boolean'),],
    #                      ids={1: ['ID']})
    # Relvar.insert(db=ev, relvar='Shaft', tuples=[
    #     Shaft_i(ID='S1', In_service=True),
    #     Shaft_i(ID='S2', In_service=False),
    # ])

    # shaft_r = Relation.restrict(db=ev, relation="shafts_rv")

    Database.close_session(ev)

    pass



"""
database.py - Create and manage a set of tclRAL databases
"""

import logging.config
from tkinter import Tcl, Tk, TclError
from pyral.exceptions import PyRALException, TclRALException
from pathlib import Path

_logger = logging.getLogger(__name__)


class Database:
    """
    Manage a set of PyRAL sessions.

    A PyRAL session is simply a tcl interpreter object that has been initialized
    by importing Andrew Mangogna's TclRAL packages.

    These packages must have been compiled for your platform and placed in the directory
    assigned to the ral_lib_path below. Consult Andrew's Model Realization repository for
    the latest files.  The most recent link should appear in the github wiki for the PyRAL
    repository or in the readme.

    A PyRAL session can then be operated on with PyRAL commands that ultimately resolve into
    TclRAL commands evalued by the associated tcl interpreter object.

    A unique human-readable name is associated with each PyRAL session so that it can be easily
    managed in a dictionary.
    """
    # TODO: Convert this class to instance methods and a singleton pattern
    # Path to the TclRAL library
    ral_lib_path = Path(__file__).parent / "tcl_scripts" / "init_TclRAL.tcl"
    sessions = {} # A dictionary of open TclRAL sessions keyed by session name
    # Temporary relational variable names keyed by session name
    # Key is the name of the owner and value is the variable name
    rv_names : dict[str, dict[str, set[str]]] = {}

    @classmethod
    def get_open_sessions(cls) -> set[str]:
        """
        Return the tcl interpreter dictionary representing all open database sessions.
        The purpose is mostly for diagnostics so that we can verify that there aren't any unneeded
        db sessions still hanging around after a procedure in the client completes.

        :param db: Name of the database session
        :return: Set of open database session names
        """
        return {k for k in cls.sessions.keys()}

    @classmethod
    def get_rv_names(cls, db: str) -> dict[str, set[str]]:
        """
        Return the relational variable names dictionary for a given database session.
        The purpose is mostly for diagnostics so that we can verify that there aren't any unneeded
        relation variables still hanging around after a procedure in the client completes.

        :param db: Name of the database session
        :return: Dictionary mapping owner -> set of RV names
        """
        # We copy the value since when we do diagnostics we might want to capture a before/after pair of values
        # and without the shallow copy(), both before and after will point to the same possibly mutated value.
        return cls.rv_names.get(db, {}).copy()

    @classmethod
    def get_all_rv_names(cls) -> dict[str, dict[str, set[str]]]:
        """
        Return the relational variable names dictionary for all open database sessions.
        The purpose is mostly for diagnostics so that we can verify that there aren't any unneeded
        relation variables still hanging around after a procedure in the client completes.

        :return: Dictionary mapping db session name -> (owner -> set of RV names)
        """
        return {db: cls.get_rv_names(db) for db in cls.get_open_sessions()}

    @classmethod
    def open_session(cls, name: str) -> Tk:
        """
        Open a PyRAL session.

        We do this by getting a new tcl interpreter and having it load TclRAL.

        Then we add the TclRAL'd tcl interpreter object to our dictionary of open sessions.

        We return the newly initiated tcl interpreter object though the client can more conveniently
        specify its session via the name.

        :param name:
        :return:  The TclRAL tcl interpreter instance is returned
        """
        # Verify that the name is not an empty string
        if not name:
            _logger.error(f"Empty '' name provided for PyRAL session")
            raise PyRALException

        # Verify that this session is not already open
        if name in cls.sessions:
            _logger.error(f"PyRAL session [{name}] already open")
            raise PyRALException

        tcl_int = Tcl()  # Get a new tcl interpreter
        tcl_int.eval(f"source {cls.ral_lib_path}")  # Load TclRAL library
        cls.sessions[name] = tcl_int  # Add it to the open session dictionary
        cls.rv_names[name] = {}  # Initialize an empty relational variable name dict for the session
        _logger.info(f"PyRAL session [{name}] initiated")

        return tcl_int

    @classmethod
    def close_session(cls, name: str):
        """
        Closes an open TclRAL session and removes any administrative data for that session

        :param name:  Session name
        """

        # Verify that this session is open
        if name not in cls.sessions:
            _logger.error(f"PyRAL session [{name}] is not open")
            raise PyRALException

        # Clear out any administrative entries related to this session
        # from class variables
        cls.sessions.pop(name, None)  # Open database session name and interpreter
        cls.rv_names.pop(name, None)  # All tracked relation variable names

        _logger.info(f"PyRAL session [{name}] closed")

    @classmethod
    def save(cls, db: str, fname: str):
        """
        Save the db in the supplied file

        :param db:  Database name
        :param fname: File name
        """
        try:
            cls.sessions[db].eval(f"serializeToFile {fname}")
        except KeyError:
            _logger.error(f"Session [{db}] not open")
            raise PyRALException

    @classmethod
    def load(cls, db: str, fname: str):
        """
        Load the db from the supplied file

        :param db:  Database name
        :param fname: File name
        """
        try:
            cls.sessions[db].eval(f"deserializeFromFile {fname}")
        except KeyError:
            _logger.error(f"Session [{db}] not open")
            raise PyRALException

    @classmethod
    def execute(cls, db: str, cmd: str, log: bool = True) -> str:
        """
        Executes a TclRAL command via the supplied session and returns TclRAL's string result.

        :param db: The DB session
        :param cmd: A TclRAL command string
        :param log:  If false, the result will not be logged. Useful when no meaningful result is expected
        :return: The string received as a result of executing the command
        """
        _logger.info(f"cmd: {cmd}")
        try:
            result = cls.sessions[db].eval(cmd)
        except KeyError:
            _logger.error(f"Database {db} is not open")
            raise PyRALException
        except TclError as e:
            _logger.error(f"TclRAL error in db [{db}] on command: [{cmd}]")
            _logger.exception(e)
            raise TclRALException

        if log:
            _logger.info(f"result: {result}")
        return result

    @classmethod
    def names(cls, db: str, pattern: str = "") -> str:
        """
        Use this to obtain names of all created relvars using the optional pattern.

        :param db:  Database name
        :param pattern:  Apply this optional pattern
        :return: TclRAl returned string
        """
        try:
            result = cls.sessions[db].eval(f"relvar names {pattern}")
        except KeyError:
            _logger.error(f"Session [{db}] not open")
            raise PyRALException

        _logger.info(f"Names in sesssion [{db}] using pattern [{pattern}]")
        _logger.info(result)
        return result

    @classmethod
    def constraint_names(cls, db: str, pattern: str = ""):
        """
        Use this to obtain names of all created constraints and names using the optional pattern.

        :param db:  Database name
        :param pattern:  Apply this optional pattern
        """
        try:
            result = cls.sessions[db].eval(f"relvar constraint names {pattern}")
        except KeyError:
            _logger.error(f"Session [{db}] not open")
            raise PyRALException

        _logger.info(f"Constraints and names in sesssion [{db}] using pattern [{pattern}]")
        _logger.info(result)
        return result



"""
join_play.py -- Play around with join

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')


acdb = "ac"  # Flow database example
def play():
    Database.open_session(acdb)
    Relvar.create_relvar(db=acdb, name='Fixed Wing Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Compass heading', 'int')], ids={1: ['ID']})
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(db=acdb, name='Pilot', attrs=[Attribute('Callsign', 'string'), Attribute('Tail_number', 'string'),
                                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])


    result = Relation.join(db=acdb, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail number": "ID"}, svar_name="join")
    result = Relation.semijoin(db=acdb, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail_number": "ID"}, svar_name="semijoin")

    Relation.print(db=acdb, variable_name="Pilot")
    Relation.print(db=acdb, variable_name="Fixed Wing Aircraft")
    Relation.print(db=acdb, variable_name="join")
    Relation.print(db=acdb, variable_name="semijoin")
    pass



""" test_sum_tblock.py -- Test summarization """

# System
import pytest
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute, SumExpr, Cardinality, RelationValue

Region_i = namedtuple('Region_i', 'Data_box Title_block_pattern Stack_order')

db = "tdb"  # Title block database example


@pytest.fixture(scope='module')
def open_db_session():
    Database.open_session(db)
    yield
    # don't close here — it's handled in tear_down


@pytest.fixture(autouse=True)
def tear_down(open_db_session):
    yield
    Database.close_session(db)


@pytest.fixture(scope='module')
def titleblock_db(open_db_session):
    Relvar.create_relvar(db=db, name='Region', attrs=[
        Attribute('Data_box', 'int'),
        Attribute('Title_block_pattern', 'string'),
        Attribute('Stack_order', 'int'),
    ], ids={1: ['Data_box', 'Title_block_pattern', 'Stack_order']})

    Relvar.insert(db=db, relvar='Region', tuples=[
        Region_i(Data_box=3, Title_block_pattern="Complex", Stack_order=1),
        Region_i(Data_box=3, Title_block_pattern="Complex", Stack_order=2),
        Region_i(Data_box=3, Title_block_pattern="SE Simple", Stack_order=1),
        Region_i(Data_box=5, Title_block_pattern="SE Simple", Stack_order=1),
        Region_i(Data_box=6, Title_block_pattern="SE Simple", Stack_order=1),
        Region_i(Data_box=6, Title_block_pattern="SE Simple", Stack_order=2),
        Region_i(Data_box=7, Title_block_pattern="SE Simple", Stack_order=1),
        Region_i(Data_box=7, Title_block_pattern="SE Simple", Stack_order=2),
    ])


def test_sum_card(titleblock_db):
    result = Relation.summarize(db=db, relation="Region", per_attrs=("Data_box", "Title_block_pattern"),
                                summaries=(SumExpr(attr=Attribute(name="Qty", type="int"), expr=Cardinality),),
                                svar_name="solution")

    expected = RelationValue(name='^relation',
                             header={'Data_box': 'int', 'Title_block_pattern': 'string', 'Qty': 'int'},
                             body=[
                                 {'Data_box': '3', 'Title_block_pattern': 'Complex', 'Qty': '2'},
                                 {'Data_box': '3', 'Title_block_pattern': 'SE Simple', 'Qty': '1'},
                                 {'Data_box': '5', 'Title_block_pattern': 'SE Simple', 'Qty': '1'},
                                 {'Data_box': '6', 'Title_block_pattern': 'SE Simple', 'Qty': '2'},
                                 {'Data_box': '7', 'Title_block_pattern': 'SE Simple', 'Qty': '2'}
                             ])
    assert result == expected



""" test_sum_flow.py -- Test summarization """

# System
import pytest
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute, SumExpr, RelationValue, SetOp, JoinCmd, SetCompareCmd, ProjectCmd

Flow_Dependency_i = namedtuple('Flow_Dependency_i', 'From_action To_action')
Actionf_i = namedtuple('Actionf_i', 'From_action')
Actionta_i = namedtuple('Actionta_i', 'To_action')
Action_i = namedtuple('Action_i', 'ID')

db = "fdb"  # Flow database example


@pytest.fixture(scope='module')
def open_db_session():
    Database.open_session(db)
    yield
    # don't close here — it's handled in tear_down


@pytest.fixture(autouse=True)
def tear_down(open_db_session):
    yield
    Database.close_session(db)


@pytest.fixture(scope='module')
def flow_db(open_db_session):
    Relvar.create_relvar(db=db, name='Flow_Dependency', attrs=[
        Attribute('From_action', 'string'),
        Attribute('To_action', 'string'),
    ], ids={1: ['From_action', 'To_action']})

    Relvar.insert(db=db, relvar='Flow_Dependency', tuples=[
        Flow_Dependency_i(From_action="ACTN1", To_action="ACTN4"),
        Flow_Dependency_i(From_action="ACTN1", To_action="ACTN2"),
        Flow_Dependency_i(From_action="ACTN2", To_action="ACTN3"),
        Flow_Dependency_i(From_action="ACTN3", To_action="ACTN7"),
        Flow_Dependency_i(From_action="ACTN4", To_action="ACTN5"),
        Flow_Dependency_i(From_action="ACTN5", To_action="ACTN6"),
        Flow_Dependency_i(From_action="ACTN6", To_action="ACTN7"),
        Flow_Dependency_i(From_action="ACTN7", To_action="ACTN9"),
        Flow_Dependency_i(From_action="ACTN8", To_action="ACTN9"),
        Flow_Dependency_i(From_action="ACTN9", To_action="ACTN13"),
        Flow_Dependency_i(From_action="ACTN9", To_action="ACTN12"),
        Flow_Dependency_i(From_action="ACTN10", To_action="ACTN11"),
        Flow_Dependency_i(From_action="ACTN11", To_action="ACTN13"),
        Flow_Dependency_i(From_action="ACTN11", To_action="ACTN12"),
    ])


def test_sum_expr(flow_db):
    # Specify a set of initial from actions that have completed execution
    Relation.create(db=db, attrs=[Attribute(name="From_action", type="string")],
                    tuples=[
                        Actionf_i(From_action="ACTN1"),
                        Actionf_i(From_action="ACTN10"),
                        Actionf_i(From_action="ACTN8"),
                    ], svar_name="xactions")
    Relation.print(db=db, variable_name="xactions")

    # Find all downstream dependencies
    Relation.join(db=db, rname1="Flow_Dependency", rname2="xactions")
    Relation.project(db=db, attributes=("To_action",), svar_name="downstream")
    Relation.print(db=db, variable_name="downstream")

    # Find all required upstream inputs for the downstream actions
    Relation.join(db=db, rname1="Flow_Dependency", rname2="downstream", svar_name="required_inputs")
    Relation.print(db=db, variable_name="required_inputs")

    # Add a boolean attribute named Can_execute
    # For each downstream action (per To_action) join it back to required inputs, projecting on the From_action
    # This yields the set of upstream executed actions for that one downstream action
    # Then test to see if this set is a subset of the xactions
    # If so, the downstream action has all dependencies fulfilled and can now execute (true)
    # Otherwise, there are some required input actions that have not just executed

    sum_expr = Relation.build_expr(commands=[
        JoinCmd(rname1="s", rname2="required_inputs", attrs=None),
        ProjectCmd(attributes=("From_action",), relation=None),
        SetCompareCmd(rname2="xactions", op=SetOp.subset, rname1=None)
    ])

    result = Relation.summarize(db=db, relation="required_inputs", per_attrs=("To_action",),
                                summaries=(SumExpr(attr=Attribute(name="Can_execute", type="boolean"), expr=sum_expr),),
                                svar_name="solution")

    expected = RelationValue(name='^relation', header={'To_action': 'string', 'Can_execute': 'boolean'},
                             body=[
                                 {'To_action': 'ACTN4', 'Can_execute': '1'},
                                 {'To_action': 'ACTN2', 'Can_execute': '1'},
                                 {'To_action': 'ACTN9', 'Can_execute': '0'},
                                 {'To_action': 'ACTN11', 'Can_execute': '1'}
                             ])
    assert result == expected



"""
transaction.py -- Database transaction
"""
import logging
from pyral.exceptions import (DuplicateTransaction, NoOpenTransaction, UnNamedTransaction, PyRALException,
                              SessionNotOpen)
from pyral.database import Database

_logger = logging.getLogger(__name__)

class Transaction:
    """
    A TclRAL transaction

    """
    pending = {}

    @classmethod
    def open(cls, db: str, name: str) -> str:
        """
        Starts a new empty transaction by ensuring that there are no statements

        :param db: The DB session
        :param name: The name of this transaction, must be unique for the db
        :return: The provided name is returned to make it easy for client to use downstream
        """
        # Verify that the name is not an empty string
        if not name:
            _logger.error(f"No transaction name provided on open for db [{db}]")
            raise UnNamedTransaction

        # Add the database to the pending dictionary if it isn't already registered
        if db not in cls.pending:
            if db not in Database.sessions:
                _logger.error(f"No open session for database [{db}]")
                raise SessionNotOpen
            cls.pending[db] = {}
        elif name in db:
            # Verify that the transaction is not already open
            _logger.error(f"Transaction {name} already pending for db [{db}]")
            raise DuplicateTransaction

        # Create a new empty tranaction
        cls.pending[db][name] = []
        _logger.info(f"OPEN > {db}:{name}")
        return name

    @classmethod
    def append_statement(cls, db: str, name: str, statement: str):
        """
        Adds a statement to the list of pending statements in the open transaction.

        :param db: The DB session
        :param name: The name of the transaction
        :param statement:  Statement to be appended
        """
        # Ensure that the statement is not empty
        if not statement:
            _logger.error(f"No statement provided on append to transaction [{name}] for db [{db}]")
            raise PyRALException

        # Verify that the transaction is pending
        if db not in cls.pending:
            _logger.error(f"No open db session: [{db}]")
            raise PyRALException

        # Append the statement to the pending transaction
        try:
            cls.pending[db][name].append(statement)
        except KeyError:
            _logger.error(f"No transaction [{name}] open on db [{db}]")
            raise NoOpenTransaction


    @classmethod
    def execute(cls, db: str, name: str):
        """
        Executes all statements in the specified transaction as a TclRAL relvar eval transaction

        :param db: The DB session
        :param name: The name of the transaction
        :return:  The TclRal success/fail result
        """
        try:
            cmd = f"relvar eval " + "{\n    " + '\n    '.join(cls.pending[db][name]) + "\n}"
        except KeyError:
            _logger.error(f"No transaction [{name}] open on db [{db}]")
            raise NoOpenTransaction

        _logger.info(f"Executing transaction:")
        _logger.info(cmd)
        result = Database.sessions[db].eval(cmd)

        # Delete the executed transaction
        del cls.pending[db][name]

        _logger.info(f"EXECUTED > {db}:{name}")
        _logger.info(f"With result: [{result}]")



"""
union_play.py -- Play around with union

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute

Aircraft_i = namedtuple('Aircraft_i', 'ID Altitude Heading')


aircraft_db = "ac"  # Flow database example
def play():
    Database.open_session(aircraft_db)
    Relvar.create_relvar(db=aircraft_db, name='Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Heading', 'int')], ids={1: ['ID']})
    Relvar.insert(db=aircraft_db, relvar='Aircraft', tuples=[
        Aircraft_i(ID='N1397Q', Altitude=13275, Heading=320),
        Aircraft_i(ID='N1309Z', Altitude=10100, Heading=273),
        Aircraft_i(ID='N5130B', Altitude=8159, Heading=90),
    ])
    R = f"ID:<N1397Q>"
    Relation.restrict(db=aircraft_db, relation='Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="a")
    R = f"ID:<N1309Z>"
    Relation.restrict(db=aircraft_db, relation='Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="b")
    R = f"ID:<N5130B>"
    Relation.restrict(db=aircraft_db, relation='Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="c")
    Relation.union(db=aircraft_db, relations=("a", "b", "c"), svar_name="u")

    c = Relation.cardinality(db=aircraft_db)

    Relation.print(db=aircraft_db, variable_name="u")
    pass



"""
rv_play.py -- Test relational variable management

"""
# System
from collections import namedtuple
from typing import NamedTuple, Any

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute

def declare_rvs(db: str, owner: str, *names: str) -> Any:
    """
    Declare multiple relation variables and return them as a dynamically named NamedTuple.
    The field names will be the declared names with '_rv' appended.
    """
    fields = [f"{name}" for name in names]
    values = [Relation.declare_rv(db=db, owner=owner, name=name) for name in names]

    RVDynamic = namedtuple("RVDynamic", fields)
    return RVDynamic(*values)


# See comment in scalar_switch.py
class RVs(NamedTuple):
    join_example: str
    semijoin_example: str

# This wrapper calls the imported declare_rvs function to generate a NamedTuple instance with each of our
# variables above as a member.
def declare_my_module_rvs(db: str, owner: str) -> RVs:
    rvs = declare_rvs(db, owner, "join_example", "semijoin_example")
    return RVs(*rvs)

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')


acdb = "ac"  # Flow database example
def play():

    Database.open_session(acdb)
    db_open = Database.get_open_sessions()


    # Get a NamedTuple with a field for each relation variable name
    rv = declare_my_module_rvs(db=acdb, owner="P1")

    Relvar.create_relvar(db=acdb, name='Fixed Wing Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Compass heading', 'int')], ids={1: ['ID']})
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(db=acdb, name='Pilot', attrs=[Attribute('Callsign', 'string'), Attribute('Tail_number', 'string'),
                                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])

    # join_example = Relation.declare_rv(db=acdb, owner="P1", name="join")
    # semi_join_example = Relation.declare_rv(db=acdb, owner="P1", name="semijoin")

    result1 = Relation.join(db=acdb, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail number": "ID"},
                            svar_name=rv.join_example)
    result2 = Relation.semijoin(db=acdb, rname1="Pilot", rname2="Fixed Wing Aircraft", attrs={"Tail_number": "ID"},
                                svar_name=rv.semijoin_example)

    Relation.print(db=acdb, variable_name=rv.join_example)
    Relation.print(db=acdb, variable_name=rv.semijoin_example)

    before = Database.get_rv_names(db=acdb)
    print("---")
    Relation.free_rvs(db=acdb, owner="P1")
    after = Database.get_rv_names(db=acdb)

    # Relation.free_rvs(db=acdb, owner="P1", names=("join_example",), exclude=True)
    Database.close_session(acdb)
    afterclose = Database.get_rv_names(db=acdb)
    afterdb = Database.get_open_sessions()

    pass



# print tests

import pytest
from pyral.relation import Relation
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.rtypes import Attribute, RelationValue

from collections import namedtuple

Fixed_Wing_Aircraft_i = namedtuple('Fixed_Wing_Aircraft_i', 'ID Altitude Compass_heading')
Pilot_i = namedtuple('Pilot_i', 'Callsign Tail_number Age')

@pytest.fixture(scope='module', autouse=True)
def tear_down():
    yield
    Database.close_session("ac")

@pytest.fixture(scope='module')
def aircraft_db():
    acdb = "ac"

    Database.open_session(acdb)
    Relvar.create_relvar(db=acdb, name='Fixed Wing Aircraft', attrs=[Attribute('ID', 'string'), Attribute('Altitude', 'int'),
                                                       Attribute('Compass heading', 'int')], ids={1: ['ID']})
    Relvar.insert(db=acdb, relvar='Fixed Wing Aircraft', tuples=[
        Fixed_Wing_Aircraft_i(ID='N1397Q', Altitude=13275, Compass_heading=320),
        Fixed_Wing_Aircraft_i(ID='N1309Z', Altitude=10100, Compass_heading=273),
        Fixed_Wing_Aircraft_i(ID='N5130B', Altitude=8159, Compass_heading=90),
    ])

    Relvar.create_relvar(db=acdb, name='Pilot', attrs=[Attribute('Callsign', 'string'), Attribute('Tail_number', 'string'),
                                                    Attribute('Age', 'int')], ids={1: ['Callsign']})
    Relvar.insert(db=acdb, relvar='Pilot', tuples=[
        Pilot_i(Callsign='Viper', Tail_number='N1397Q', Age=22),
        Pilot_i(Callsign='Joker', Tail_number='N5130B', Age=31),
    ])
    return acdb


def test_compare_equal(aircraft_db):
    result = Relation.compare(db=aircraft_db, op='==', rname1='Fixed Wing Aircraft', rname2='Fixed Wing Aircraft')
    expected = True
    assert result == expected


def test_compare_not_equal(aircraft_db):
    result = Relation.compare(db=aircraft_db, op='!=', rname1='Fixed Wing Aircraft', rname2='Fixed Wing Aircraft')
    expected = False
    assert result == expected


def test_intersect(aircraft_db):
    cmd_high = 'set high [relation restrict $Fixed_Wing_Aircraft t {[tuple extract $t Altitude] > 9000}]'
    cmd_low = 'set low [relation restrict $Fixed_Wing_Aircraft t {[tuple extract $t Altitude] < 13000}]'
    Database.execute(db=aircraft_db, cmd=cmd_high)
    Database.execute(db=aircraft_db, cmd=cmd_low)
    Relation.print(db=aircraft_db, variable_name='high')
    Relation.print(db=aircraft_db, variable_name='low')
    b = Relation.intersect(db=aircraft_db, rname2='high', rname1='low')
    expected = RelationValue(name='^relation',
                             header={'ID': 'string', 'Altitude': 'int', 'Compass_heading': 'int'},
                             body=[{'ID': 'N1309Z', 'Altitude': '10100', 'Compass_heading': '273'}])
    assert b == expected
    Relation.relformat(b)

def test_cardinality(aircraft_db):
    c = Relation.cardinality(db=aircraft_db, rname="Fixed Wing Aircraft")
    assert c == 3

def test_union(aircraft_db):
    R = f"ID:<N1397Q>"
    Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="a")
    R = f"ID:<N1309Z>"
    Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="b")
    R = f"ID:<N5130B>"
    Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction=R)
    Relation.project(db=aircraft_db, attributes=("ID",), svar_name="c")
    u = Relation.union(db=aircraft_db, relations=("a", "b", "c"))

    expected = RelationValue(name='^relation',
                             header={'ID': 'string'},
                             body=[{'ID': 'N1397Q'}, {'ID': 'N1309Z'}, {'ID': 'N5130B'}])
    Relation.relformat(u)
    assert u == expected


def test_join(aircraft_db):
    result = Relation.join(db=aircraft_db, rname2='Fixed Wing Aircraft', rname1='Pilot',
                           attrs={'Tail_number': 'ID'}, svar_name='Joined')
    expected = RelationValue(name='^relation',
                             header={'Callsign': 'string', 'Tail_number': 'string', 'Age': 'int', 'Altitude': 'int',
                                     'Compass_heading': 'int'},
                             body=[{'Callsign': 'Viper', 'Tail_number': 'N1397Q', 'Age': '22', 'Altitude': '13275',
                                    'Compass_heading': '320'},
                                   {'Callsign': 'Joker', 'Tail_number': 'N5130B', 'Age': '31', 'Altitude': '8159',
                                    'Compass_heading': '90'}])
    Relation.relformat(result)
    assert result == expected

def test_semijoin(aircraft_db):
    result = Relation.semijoin(db=aircraft_db, rname1='Pilot', rname2='Fixed Wing Aircraft', attrs={'Tail number': 'ID'})
    expected = RelationValue(name='^relation',
                             header={'ID': 'string', 'Altitude': 'int', 'Compass_heading': 'int'},
                             body=[{'ID': 'N1397Q', 'Altitude': '13275', 'Compass_heading': '320'},
                                   {'ID': 'N5130B', 'Altitude': '8159', 'Compass_heading': '90'}])
    Relation.relformat(result)
    assert result == expected


def test_selectid_found(aircraft_db):
    result = Relvar.select_id(db=aircraft_db, relvar_name='Fixed Wing Aircraft', tid={'ID': 'N1397Q'})
    expected = RelationValue(name=None, header={'ID': 'string', 'Altitude': 'int', 'Compass_heading': 'int'},
                             body=[{'ID': 'N1397Q', 'Altitude': '13275', 'Compass_heading': '320'}])
    Relation.relformat(result)
    assert result == expected


def test_selectid_none(aircraft_db):
    result = Relvar.select_id(db=aircraft_db, relvar_name='Fixed Wing Aircraft', tid={'ID': 'X'})
    expected = RelationValue(name=None, header={'ID': 'string', 'Altitude': 'int', 'Compass_heading': 'int'},
                             body={})
    assert result == expected


def test_restrict(aircraft_db):
    R = f"ID:<N1397Q>"
    result = Relation.restrict(db=aircraft_db, relation='Fixed Wing Aircraft', restriction=R)
    expected = RelationValue(name='^relation', header={'ID': 'string', 'Altitude': 'int', 'Compass_heading': 'int'},
                             body=[{'ID': 'N1397Q', 'Altitude': '13275', 'Compass_heading': '320'}])
    assert result == expected



"""
bool_play.py -- Test bool data type in Python and TclRAL

"""
# System
from collections import namedtuple

# PyRAL
from pyral.database import Database
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.rtypes import Attribute

Shaft_i = namedtuple('Shaft_i', 'ID In_service')

ev = "ev"

def play():
    Database.open_session(ev)
    Relation.create(db=ev, attrs=[Attribute(name="ID", type="string"), Attribute(name="In_service", type="boolean")],
                    tuples=[
                        Shaft_i(ID='S1', In_service=True),
                        Shaft_i(ID='S2', In_service=False),
                        Shaft_i(ID='S3', In_service=True),
                    ], svar_name="shafts_rv")
    # Relation.raw(db=ev, cmd_str="relation body $shafts_rv")

    R = f"In_service:<{True}>"
    result = Relation.restrict(db=ev, relation="shafts_rv", restriction=R)

    pass

    # Relvar.create_relvar(db=ev, name='Shaft', attrs=[Attribute('ID', 'string'), Attribute('In_service', 'boolean'),],
    #                      ids={1: ['ID']})
    # Relvar.insert(db=ev, relvar='Shaft', tuples=[
    #     Shaft_i(ID='S1', In_service=True),
    #     Shaft_i(ID='S2', In_service=False),
    # ])

    shaft_r = Relation.restrict(db=ev, relation="shafts_rv")



    Relation.print(db=ev, variable_name="shafts_rv")

    Database.close_session(ev)

    pass
